from dataclasses import dataclass

from textual import containers
from textual.binding import Binding
from textual import events
from textual.message import Message
from textual.reactive import reactive
from textual.layouts.grid import GridLayout
from textual.widget import Widget


class GridSelect(containers.ItemGrid, can_focus=True):
    FOCUS_ON_CLICK = False
    CURSOR_GROUP = Binding.Group("Select")
    FOCUS_GROUP = Binding.Group("Focus")
    BINDINGS = [
        Binding("up", "cursor_up", "Cursor Up", group=CURSOR_GROUP),
        Binding("down", "cursor_down", "Cursor Down", group=CURSOR_GROUP),
        Binding("left", "cursor_left", "Cursor Left", group=CURSOR_GROUP),
        Binding("right", "cursor_right", "Cursor Right", group=CURSOR_GROUP),
        Binding("enter", "select", "Select"),
    ]

    highlighted: reactive[int | None] = reactive(None)

    @dataclass
    class Selected(Message):
        grid_select: "GridSelect"
        selected_widget: Widget

        @property
        def control(self) -> Widget:
            return self.grid_select

    def __init__(
        self,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        min_column_width: int = 30,
        max_column_width: int | None = None,
    ):
        super().__init__(
            name=name,
            id=id,
            classes=classes,
            min_column_width=min_column_width,
            max_column_width=max_column_width,
        )

    @property
    def grid_size(self) -> tuple[int, int] | None:
        assert isinstance(self.layout, GridLayout)
        return self.layout.grid_size

    def check_action(self, action: str, parameters: tuple[object, ...]) -> bool | None:
        if action in {"cursor_up", "cursor_down", "cursor_left", "cursor_right"}:
            return (
                None
                if (
                    (self.grid_size is None or self.highlighted is None)
                    or len(self.children) <= 1
                )
                else True
            )
        return True

    def on_focus(self):
        if self.highlighted is None:
            self.highlighted = 0
        self.reveal_highlight()

    def on_blur(self) -> None:
        self.highlighted = None

    def reveal_highlight(self):
        if self.highlighted is None:
            return
        try:
            highlighted_widget = self.children[self.highlighted]
        except IndexError:
            pass
        else:
            if not self.screen.can_view_entire(highlighted_widget):
                self.screen.scroll_to_center(highlighted_widget, origin_visible=True)

    def watch_highlighted(
        self, old_highlighted: int | None, highlighted: int | None
    ) -> None:
        if old_highlighted is not None:
            try:
                self.children[old_highlighted].remove_class("-highlight")
            except IndexError:
                pass
        if highlighted is not None:
            try:
                highlighted_widget = self.children[highlighted]
                highlighted_widget.add_class("-highlight")
            except IndexError:
                pass
        self.reveal_highlight()

    def validate_highlighted(self, highlighted: int | None) -> int | None:
        if highlighted is None:
            return None

        if not self.children:
            return None
        if highlighted < 0:
            return 0
        if highlighted >= len(self.children):
            return len(self.children) - 1
        return highlighted

    def action_cursor_up(self):
        if (grid_size := self.grid_size) is None:
            return
        if self.highlighted is None:
            self.highlighted = 0
        else:
            width, _height = grid_size
            if self.highlighted >= width:
                self.highlighted -= width

    def action_cursor_down(self):
        if (grid_size := self.grid_size) is None:
            return
        if self.highlighted is None:
            self.highlighted = 0
        else:
            width, height = grid_size
            if self.highlighted + width < len(self.children):
                self.highlighted += width

    def action_cursor_left(self):
        if self.highlighted is None:
            self.highlighted = 0
        else:
            self.highlighted -= 1

    def action_cursor_right(self):
        if self.highlighted is None:
            self.highlighted = 0
        else:
            self.highlighted += 1

    def on_click(self, event: events.Click) -> None:
        if event.widget is None:
            return

        highlighted_widget: Widget | None = None
        if self.highlighted is not None:
            try:
                highlighted_widget = self.children[self.highlighted]
            except IndexError:
                pass
        for widget in event.widget.ancestors_with_self:
            if widget in self.children:
                if highlighted_widget is not None and highlighted_widget is widget:
                    self.action_select()
                else:
                    self.highlighted = self.children.index(widget)
                break
        self.focus()

    def action_select(self):
        if self.highlighted is not None:
            try:
                highlighted_widget = self.children[self.highlighted]
            except IndexError:
                pass
            else:
                self.post_message(self.Selected(self, highlighted_widget))


if __name__ == "__main__":
    from textual.app import App, ComposeResult
    from textual import widgets

    class GridApp(App):
        CSS = """
        .grid-item {
            width: 1fr;
            padding: 0 1;
            # background: blue 20%;        
            border: blank;

            &:hover {
                background: $panel;
            }

            &.-highlight {
                border: tall $primary;
                background: $panel;
            }
        }
        """

        def compose(self) -> ComposeResult:
            yield widgets.Footer()
            with GridSelect():
                for n in range(50):
                    yield widgets.Label(
                        f"#{n} Where there is a Will, there is a Way!",
                        classes="grid-item",
                    )

    GridApp().run()



from __future__ import annotations


import asyncio
from functools import lru_cache
from operator import itemgetter
from pathlib import Path
import re
from typing import Sequence

import pathspec.patterns
from pathspec import PathSpec

from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual import work
from textual import getters
from textual import containers
from textual.reactive import var, Initialize
from textual.content import Content, Span
from textual.widget import Widget
from textual.widgets import OptionList, Input
from textual.widgets.option_list import Option


from toad import directory
from toad.fuzzy import FuzzySearch
from toad.messages import Dismiss, InsertPath, PromptSuggestion


class PathFuzzySearch(FuzzySearch):
    @classmethod
    @lru_cache(maxsize=1024)
    def get_first_letters(cls, candidate: str) -> frozenset[int]:
        return frozenset(
            {
                0,
                *[match.start() + 1 for match in re.finditer(r"/", candidate)],
            }
        )

    def score(self, candidate: str, positions: Sequence[int]) -> float:
        """Score a search.

        Args:
            search: Search object.

        Returns:
            Score.
        """
        first_letters = self.get_first_letters(candidate)
        # This is a heuristic, and can be tweaked for better results
        # Boost first letter matches
        offset_count = len(positions)
        score: float = offset_count + len(first_letters.intersection(positions))

        # if 0 in first_letters:
        #     score += 1

        groups = 1
        last_offset, *offsets = positions
        for offset in offsets:
            if offset != last_offset + 1:
                groups += 1
            last_offset = offset

        # Boost to favor less groups
        normalized_groups = (offset_count - (groups - 1)) / offset_count
        score *= 1 + (normalized_groups * normalized_groups)
        return score


class PathSearch(containers.VerticalGroup):
    CURSOR_BINDING_GROUP = Binding.Group(description="Move selection")
    BINDINGS = [
        Binding(
            "up", "cursor_up", "Cursor up", group=CURSOR_BINDING_GROUP, priority=True
        ),
        Binding(
            "down",
            "cursor_down",
            "Cursor down",
            group=CURSOR_BINDING_GROUP,
            priority=True,
        ),
        Binding("enter", "submit", "Insert path", priority=True),
        Binding("escape", "dismiss", "Dismiss", priority=True),
    ]

    def get_fuzzy_search(self) -> FuzzySearch:
        return PathFuzzySearch(case_sensitive=False)

    root: var[Path] = var(Path("./"))
    paths: var[list[Path]] = var(list)
    highlighted_paths: var[list[Content]] = var(list)
    filtered_path_indices: var[list[int]] = var(list)
    loaded = var(False)
    filter = var("")
    fuzzy_search: var[FuzzySearch] = var(Initialize(get_fuzzy_search))

    option_list = getters.query_one(OptionList)
    input = getters.query_one(Input)

    def compose(self) -> ComposeResult:
        yield Input(compact=True, placeholder="fuzzy search")
        yield OptionList()

    async def search(self, search: str) -> None:
        if not search:
            self.option_list.set_options(
                [
                    Option(highlighted_path, highlighted_path.plain)
                    for highlighted_path in self.highlighted_paths
                ],
            )
            return

        fuzzy_search = self.fuzzy_search
        fuzzy_search.cache.grow(len(self.paths))
        scores: list[tuple[float, Sequence[int], Content]] = [
            (
                *fuzzy_search.match(search, highlighted_path.plain),
                highlighted_path,
            )
            for highlighted_path in self.highlighted_paths
        ]

        scores = sorted(
            [score for score in scores if score[0]], key=itemgetter(0), reverse=True
        )

        def highlight_offsets(path: Content, offsets: Sequence[int]) -> Content:
            return path.add_spans(
                [Span(offset, offset + 1, "underline") for offset in offsets]
            )

        self.option_list.set_options(
            [
                Option(highlight_offsets(path, offsets), id=path.plain)
                for score, offsets, path in scores
            ]
        )
        self.option_list.highlighted = 0
        self.post_message(PromptSuggestion(""))

    def action_cursor_down(self) -> None:
        self.option_list.action_cursor_down()

    def action_cursor_up(self) -> None:
        self.option_list.action_cursor_up()

    def action_dismiss(self) -> None:
        self.post_message(Dismiss(self))

    def focus(self, scroll_visible: bool = False) -> Self:
        return self.input.focus(scroll_visible=scroll_visible)

    @on(Input.Changed)
    async def on_input_changed(self, event: Input.Changed):
        await self.search(event.value)

    @on(OptionList.OptionHighlighted)
    async def on_option_list_changed(self, event: OptionList.OptionHighlighted):
        event.stop()
        if event.option:
            self.post_message(PromptSuggestion(event.option.id))

    @on(OptionList.OptionSelected)
    def on_option_list_option_selected(self, event: OptionList.OptionSelected) -> None:
        self.action_submit()

    def action_submit(self):
        if (highlighted := self.option_list.highlighted) is not None:
            option = self.option_list.options[highlighted]
            if option.id:
                self.post_message(InsertPath(option.id))
                self.post_message(Dismiss(self))

    def watch_root(self, root: Path) -> None:
        pass

    def get_path_spec(self, git_ignore_path: Path) -> PathSpec | None:
        """Get a path spec instance if there is a .gitignore file present.

        Args:
            git_ignore_path): Path to .gitignore.

        Returns:
            A `PathSpec` instance.
        """
        try:
            if git_ignore_path.is_file():
                spec_text = git_ignore_path.read_text()
                spec = PathSpec.from_lines(
                    pathspec.patterns.GitWildMatchPattern, spec_text.splitlines()
                )
                return spec
        except OSError:
            return None
        return None

    @work(exclusive=True)
    async def load_paths(self) -> None:
        self.input.clear()
        self.input.focus()
        root = self.root

        self.loading = True

        path_spec = await asyncio.to_thread(self.get_path_spec, root / ".gitignore")
        # path_spec = await self.get_path_spec(root / ".gitignore").wait()
        paths = await directory.scan(root, path_spec=path_spec, add_directories=True)

        paths = [path.absolute() for path in paths]
        # paths.sort(key=lambda path: (len(path.parts), str(path).lower()))
        self.root = root
        self.paths = paths
        self.loading = False

    def get_loading_widget(self) -> Widget:
        from textual.widgets import LoadingIndicator

        return LoadingIndicator()

    def highlight_path(self, path: str) -> Content:
        content = Content.styled(path, "dim")
        if path.startswith("."):
            return content
        if not path.endswith("/"):
            if "/" in path:
                content = content.stylize("$text-success", path.rfind("/") + 1)
            else:
                content = content.stylize("$text-success dim")
        if (match := re.search(r"\.(.*$)", content.plain)) is not None:
            content = content.stylize("not dim", match.start(1), match.end(1))
        return content

    def watch_paths(self, paths: list[Path]) -> None:
        self.option_list.highlighted = None

        def path_display(path: Path) -> str:
            try:
                is_directory = path.is_dir()
            except OSError:
                is_directory = False
            if is_directory:
                return str(path.relative_to(self.root)) + "/"
            else:
                return str(path.relative_to(self.root))

        display_paths = sorted(map(path_display, paths), key=str.lower)
        self.highlighted_paths = [self.highlight_path(path) for path in display_paths]
        self.option_list.set_options(
            [
                Option(highlighted_path, id=highlighted_path.plain)
                for highlighted_path in self.highlighted_paths
            ],
        )
        self.option_list.highlighted = 0
        self.post_message(PromptSuggestion(""))
        self.input.focus()



from __future__ import annotations


from contextlib import suppress
import os
import asyncio
import codecs
import fcntl
import platform
import pty
import struct
import termios
from dataclasses import dataclass
from typing import TYPE_CHECKING

from textual.message import Message

from toad.shell_read import shell_read

from toad.widgets.terminal import Terminal

if TYPE_CHECKING:
    from toad.widgets.conversation import Conversation

IS_MACOS = platform.system() == "Darwin"


def resize_pty(fd, cols, rows):
    """Resize the pseudo-terminal"""
    # Pack the dimensions into the format expected by TIOCSWINSZ
    try:
        size = struct.pack("HHHH", rows, cols, 0, 0)
        fcntl.ioctl(fd, termios.TIOCSWINSZ, size)
    except OSError:
        # Possibly file descriptor closed
        pass


@dataclass
class CurrentWorkingDirectoryChanged(Message):
    """Current working directory has changed in shell."""

    path: str


@dataclass
class ShellFinished(Message):
    """The shell finished."""


class Shell:
    """Responsible for shell interactions in Conversation."""

    def __init__(
        self,
        conversation: Conversation,
        working_directory: str,
        shell="",
        start="",
    ) -> None:
        self.conversation = conversation
        self.working_directory = working_directory

        self.terminal: Terminal | None = None
        self.new_log: bool = False
        self.shell = shell or os.environ.get("SHELL", "sh")
        self.shell_start = start
        self.master: int | None = None
        self._task: asyncio.Task | None = None
        self._process: asyncio.subprocess.Process | None = None

        self._finished: bool = False
        self._ready_event: asyncio.Event = asyncio.Event()

        self._hide_echo: set[bytes] = set()
        """A set of byte strings to remove from output."""

    @property
    def is_finished(self) -> bool:
        return self._finished

    async def send(self, command: str, width: int, height: int) -> None:
        await self._ready_event.wait()
        if self.master is None:
            print("TTY FD not set")
            return

        self.terminal = None
        await asyncio.to_thread(resize_pty, self.master, width, max(height, 1))

        get_pwd_command = f"{command};" + r'printf "\e]2025;$(pwd);\e\\"' + "\n"
        await self.write(get_pwd_command, hide_echo=True)

    def start(self) -> None:
        assert self._task is None
        self._task = asyncio.create_task(self.run(), name=repr(self))

    async def interrupt(self) -> None:
        """Interrupt the running command."""
        await self.write(b"\x03")

    def update_size(self, width: int, height: int) -> None:
        """Update the size of the shell pty.

        Args:
            width: Desired width.
            height: Desired height.
        """
        if self.master is None:
            return
        with suppress(OSError):
            resize_pty(self.master, width, max(height, 1))

    async def write(self, text: str | bytes, hide_echo: bool = False) -> int:
        if self.master is None:
            return 0
        text_bytes = text.encode("utf-8", "ignore") if isinstance(text, str) else text
        if hide_echo:
            self._hide_echo.add(text_bytes)
        return await asyncio.to_thread(os.write, self.master, text_bytes)

    async def run(self) -> None:
        current_directory = self.working_directory

        master, slave = pty.openpty()
        self.master = master

        flags = fcntl.fcntl(master, fcntl.F_GETFL)
        fcntl.fcntl(master, fcntl.F_SETFL, flags | os.O_NONBLOCK)

        env = os.environ.copy()
        env["FORCE_COLOR"] = "1"
        env["TTY_COMPATIBLE"] = "1"
        env["TERM"] = "xterm-256color"
        env["COLORTERM"] = "truecolor"
        env["TOAD"] = "1"
        env["CLICOLOR"] = "1"

        shell = self.shell

        def setup_pty():
            os.setsid()
            fcntl.ioctl(slave, termios.TIOCSCTTY, 0)

        try:
            _process = await asyncio.create_subprocess_shell(
                shell,
                stdin=slave,
                stdout=slave,
                stderr=slave,
                env=env,
                cwd=current_directory,
                preexec_fn=setup_pty,
            )
        except Exception as error:
            self.conversation.notify(
                f"Unable to start shell: {error}\n\nCheck your settings.",
                title="Shell",
                severity="error",
            )
            return

        os.close(slave)
        BUFFER_SIZE = 64 * 1024
        reader = asyncio.StreamReader(BUFFER_SIZE)
        protocol = asyncio.StreamReaderProtocol(reader)

        loop = asyncio.get_event_loop()
        transport, _ = await loop.connect_read_pipe(
            lambda: protocol, os.fdopen(master, "rb", 0)
        )

        self._ready_event.set()

        if shell_start := self.shell_start.strip():
            shell_start = self.shell_start.strip()
            if not shell_start.endswith("\n"):
                shell_start += "\n"
            await self.write(shell_start, hide_echo=True)

        unicode_decoder = codecs.getincrementaldecoder("utf-8")(errors="replace")

        while True:
            data = await shell_read(reader, BUFFER_SIZE)

            for string_bytes in list(self._hide_echo):
                remove_bytes = string_bytes.replace(b"\n", b"\r\n")
                if remove_bytes in data:
                    data = data.replace(remove_bytes, b"")
                    self._hide_echo.discard(string_bytes)
                    if not data:
                        data = b"\r"

            if line := unicode_decoder.decode(data, final=not data):
                if self.terminal is None or self.terminal.is_finalized:
                    previous_state = (
                        None if self.terminal is None else self.terminal.state
                    )
                    self.terminal = await self.conversation.new_terminal()
                    # if previous_state is not None:
                    #     self.terminal.set_state(previous_state)
                    self.terminal.set_write_to_stdin(self.write)
                if await self.terminal.write(line) and not self.terminal.display:
                    if (
                        self.terminal.alternate_screen
                        or not self.terminal.state.scrollback_buffer.is_blank
                    ):
                        self.terminal.display = True
                new_directory = self.terminal.current_directory
                if new_directory and new_directory != current_directory:
                    current_directory = new_directory
                    self.conversation.post_message(
                        CurrentWorkingDirectoryChanged(current_directory)
                    )
            if (
                self.terminal is not None
                and self.terminal.is_finalized
                and self.terminal.state.scrollback_buffer.is_blank
            ):
                await self.terminal.remove()
                self.terminal = None

            if not data:
                break

        self.master = None
        self._finished = True
        self.conversation.post_message(ShellFinished())



"""
Fuzzy matcher.

This class is used by the [command palette](/guide/command_palette) to match search terms.

"""

from __future__ import annotations

from functools import lru_cache
from operator import itemgetter
from re import finditer
from typing import Iterable, Sequence

import rich.repr

from textual.cache import LRUCache
from textual.content import Content
from textual.visual import Style


class FuzzySearch:
    """Performs a fuzzy search.

    Unlike a regex solution, this will finds all possible matches.
    """

    def __init__(
        self, case_sensitive: bool = False, *, cache_size: int = 1024 * 4
    ) -> None:
        """Initialize fuzzy search.

        Args:
            case_sensitive: Is the match case sensitive?
            cache_size: Number of queries to cache.
        """

        self.case_sensitive = case_sensitive
        self.cache: LRUCache[tuple[str, str], tuple[float, Sequence[int]]] = LRUCache(
            cache_size
        )

    def match(self, query: str, candidate: str) -> tuple[float, Sequence[int]]:
        """Match against a query.

        Args:
            query: The fuzzy query.
            candidate: A candidate to check,.

        Returns:
            A pair of (score, tuple of offsets). `(0, ())` for no result.
        """

        cache_key = (query, candidate)
        if cache_key in self.cache:
            return self.cache[cache_key]
        default: tuple[float, Sequence[int]] = (0.0, [])
        result = max(self._match(query, candidate), key=itemgetter(0), default=default)
        self.cache[cache_key] = result
        return result

    @classmethod
    @lru_cache(maxsize=1024)
    def get_first_letters(cls, candidate: str) -> frozenset[int]:
        return frozenset({match.start() for match in finditer(r"\w+", candidate)})

    def score(self, candidate: str, positions: Sequence[int]) -> float:
        """Score a search.

        Args:
            search: Search object.

        Returns:
            Score.
        """
        first_letters = self.get_first_letters(candidate)
        # This is a heuristic, and can be tweaked for better results
        # Boost first letter matches
        offset_count = len(positions)
        score: float = offset_count + len(first_letters.intersection(positions))

        groups = 1
        last_offset, *offsets = positions
        for offset in offsets:
            if offset != last_offset + 1:
                groups += 1
            last_offset = offset

        # Boost to favor less groups
        normalized_groups = (offset_count - (groups - 1)) / offset_count
        score *= 1 + (normalized_groups * normalized_groups)
        return score

    def _match(
        self, query: str, candidate: str
    ) -> Iterable[tuple[float, Sequence[int]]]:
        letter_positions: list[list[int]] = []
        position = 0

        if not self.case_sensitive:
            candidate = candidate.lower()
            query = query.lower()

        score = self.score

        for offset, letter in enumerate(query):
            last_index = len(candidate) - offset
            positions: list[int] = []
            letter_positions.append(positions)
            index = position
            while (location := candidate.find(letter, index)) != -1:
                positions.append(location)
                index = location + 1
                if index >= last_index:
                    break
            if not positions:
                yield (0.0, ())
                return
            position = positions[0] + 1

        possible_offsets: list[list[int]] = []
        query_length = len(query)

        def get_offsets(offsets: list[int], positions_index: int) -> None:
            """Recursively match offsets.

            Args:
                offsets: A list of offsets.
                positions_index: Index of query letter.

            """
            for offset in letter_positions[positions_index]:
                if not offsets or offset > offsets[-1]:
                    new_offsets = [*offsets, offset]
                    if len(new_offsets) == query_length:
                        possible_offsets.append(new_offsets)
                    else:
                        get_offsets(new_offsets, positions_index + 1)

        get_offsets([], 0)

        for offsets in possible_offsets:
            yield score(candidate, offsets), offsets



import shutil

from textual.app import ComposeResult
from textual import on, work
from textual import containers
from textual import getters
from textual.screen import ModalScreen
from textual import widgets
from textual.widget import Widget

from toad.app import ToadApp
from toad.widgets.command_pane import CommandPane


UV_INSTALL = "curl -LsSf https://astral.sh/uv/install.sh | sh"


class ActionModal(ModalScreen):
    """Executes an action command."""

    command_pane = getters.query_one(CommandPane)
    ok_button = getters.query_one("#ok", widgets.Button)

    app = getters.app(ToadApp)

    BINDINGS = [("escape", "dismiss_modal", "Dismiss")]

    def __init__(
        self,
        action: str,
        agent: str,
        title: str,
        command: str,
        *,
        bootstrap_uv: bool = False,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
    ) -> None:
        self._action = action
        self._agent = agent
        self._title = title
        self._command = command
        self._bootstrap_uv = bootstrap_uv
        super().__init__(name=name, id=id, classes=classes)

    def get_loading_widget(self) -> Widget:
        return widgets.LoadingIndicator()

    def compose(self) -> ComposeResult:
        with containers.VerticalGroup(id="container"):
            yield CommandPane()
            yield widgets.Button("OK", id="ok", disabled=True)

    def enable_button(self) -> None:
        self.ok_button.loading = False
        self.ok_button.disabled = False
        self.ok_button.focus()

    @on(CommandPane.CommandComplete)
    def on_command_complete(self, event: CommandPane.CommandComplete) -> None:
        self.enable_button()

    def on_mount(self) -> None:
        self.ok_button.loading = True
        self.command_pane.border_title = self._title

        self.run_command()

    @work()
    async def run_command(self) -> None:
        """Write and execute the command."""
        self.command_pane.anchor()
        if self._bootstrap_uv and shutil.which("uv") is None:
            # Bootstrap UV if required
            await self.command_pane.write(f"$ {UV_INSTALL}\n")
            await self.command_pane.execute(UV_INSTALL, final=False)

        await self.command_pane.write(f"$ {self._command}\n")
        action_task = self.command_pane.execute(self._command)
        await action_task
        self.app.capture_event(
            "agent-action",
            action=self._action,
            agent=self._agent,
            fail=self.command_pane.return_code != 0,
        )

    @on(widgets.Button.Pressed)
    def on_button_pressed(self) -> None:
        self.action_dismiss_modal()

    def action_dismiss_modal(self) -> None:
        self.dismiss(self.command_pane.return_code)



from typing import cast

from textual import on
from textual import getters
from textual.app import ComposeResult

from textual import work
from textual.screen import ModalScreen
from textual import containers
from textual import widgets
from textual.reactive import var

import toad
from textual.binding import Binding
from toad.agent_schema import Action, Agent, OS, Command
from toad.app import ToadApp


class AgentModal(ModalScreen):
    AUTO_FOCUS = "#launcher-checkbox"

    BINDINGS = [
        Binding("escape", "dismiss(None)", "Dismiss", show=False),
        Binding("space", "dismiss('launch')", "Launch agent", priority=True),
    ]

    action = var("")

    app = getters.app(ToadApp)
    action_select = getters.query_one("#action-select", widgets.Select)
    launcher_checkbox = getters.query_one("#launcher-checkbox", widgets.Checkbox)

    def __init__(self, agent: Agent) -> None:
        self._agent = agent
        super().__init__()

    def compose(self) -> ComposeResult:
        launcher_set = frozenset(
            self.app.settings.get("launcher.agents", str).splitlines()
        )

        agent = self._agent

        app = self.app
        launcher_set = frozenset(app.settings.get("launcher.agents", str).splitlines())
        agent = self._agent
        actions = agent["actions"]

        script_os = cast(OS, toad.os)
        if script_os not in actions:
            script_os = "*"

        commands: dict[Action, Command] = actions[cast(OS, script_os)]
        script_choices = [
            (action["description"], name) for name, action in commands.items()
        ]

        with containers.Vertical(id="container"):
            with containers.VerticalScroll(id="description-container"):
                yield widgets.Markdown(agent["help"], id="description")
            with containers.VerticalGroup():
                with containers.HorizontalGroup():
                    yield widgets.Checkbox(
                        "Show in launcher",
                        value=agent["identity"] in launcher_set,
                        id="launcher-checkbox",
                    )
                    yield widgets.Select(
                        script_choices,
                        prompt="Actions",
                        allow_blank=True,
                        id="action-select",
                    )
                    yield widgets.Button(
                        "Go", variant="primary", id="run-action", disabled=True
                    )
        yield widgets.Footer()

    def on_mount(self) -> None:
        self.query_one("Footer").styles.animate("opacity", 1.0, duration=500 / 1000)

    @on(widgets.Checkbox.Changed)
    def on_checkbox_changed(self, event: widgets.Select.Changed) -> None:
        launcher_agents = self.app.settings.get("launcher.agents", str).splitlines()
        agent_identity = self._agent["identity"]
        if agent_identity in launcher_agents:
            launcher_agents.remove(agent_identity)
        if event.value:
            launcher_agents.insert(0, agent_identity)
        self.app.settings.set("launcher.agents", "\n".join(launcher_agents))

    @on(widgets.Select.Changed)
    def on_select_changed(self, event: widgets.Select.Changed) -> None:
        self.action = event.value if isinstance(event.value, str) else ""

    @work
    @on(widgets.Button.Pressed)
    async def on_button_pressed(self) -> None:
        agent = self._agent
        action = self.action_select.value
        assert isinstance(action, str)
        agent_actions = self._agent["actions"]

        if (commands := agent_actions.get(toad.os, None)) is None:
            commands = agent_actions.get("*", None)
        if commands is None:
            self.notify(
                "Action is not available on this platform",
                title="Agent action",
                severity="error",
            )
            return
        command = commands[action]

        from toad.screens.action_modal import ActionModal

        title = command["description"]
        agent_id = self._agent["identity"]
        action_command = command["command"]
        bootstrap_uv = command.get("bootstrap_uv", False)

        agent = self._agent
        # Focus the select
        # It's unlikely the user wants to re-run the action
        self.action_select.focus()
        return_code = await self.app.push_screen_wait(
            ActionModal(
                action,
                agent_id,
                title,
                action_command,
                bootstrap_uv=bootstrap_uv,
            )
        )
        if return_code == 0 and action in {"install", "install-acp"}:
            # Add to launcher if we installed something
            if not self.launcher_checkbox.value:
                self.notify(
                    f"{agent['name']} has been added to your launcher",
                    title="Add agent",
                    severity="information",
                )
                self.launcher_checkbox.value = True

    def watch_action(self, action: str) -> None:
        go_button = self.query_one("#run-action", widgets.Button)
        go_button.disabled = not action
